/*
** File: linux-2.6.c  
** Project: Linux ADB3 monolithic driver
** Purpose: Entry point for kernel module and other Linux-specific declarations.
**
** (C) Copyright Alpha Data 2010-2012
*/

#include <linux/module.h>
#include <linux/init.h>

#include "df.h"
#include "dflinux.h"

static const char* g_pciDriverName = "adb3";

MODULE_LICENSE("GPL");

/*
** Module parameters
*/

static uint Admxrc6tadv8CompatSensor1Name = 0;
module_param(Admxrc6tadv8CompatSensor1Name, uint, S_IRUGO);
MODULE_PARM_DESC(
  Admxrc6tadv8CompatSensor1Name,
  "If nonzero, sensor 1 on ADM-XRC-6T-ADV8 has old, incorrect name (for compatibility with software that relies on old sensor name). Default is 0.");

static uint AssumeXrmPresent = 0;
module_param(AssumeXrmPresent, uint, S_IRUGO);
MODULE_PARM_DESC(
  AssumeXrmPresent,
  "If nonzero, assumes that an XRM I/O personality module is present (for XRMs that do not correctly assert presence detect signal). Default is 0.");

static uint DebugLevel = 0;
module_param(DebugLevel, uint, S_IRUGO);
MODULE_PARM_DESC(
  DebugLevel,
  "Verbosity of debug messages to system log; 0 is minimum and higher is more verbose. Default is 0.");

static uint DmaMapLength = 64;
module_param(DmaMapLength, uint, S_IRUGO);
MODULE_PARM_DESC(
  DmaMapLength,
  "Maximum number of user-space pages to map in a single chunk of a DMA operation. Specifying too large a value may result in the driver failing to start due to memory allocation failures. Default is 64.");

static uint EnableVpdWrite = 0;
module_param(EnableVpdWrite, uint, S_IRUGO);
MODULE_PARM_DESC(
  EnableVpdWrite,
  "If nonzero, permits writes to VPD memory. Default is 0.");

static uint IncludeDiagSensors = 0;
module_param(IncludeDiagSensors, uint, S_IRUGO);
MODULE_PARM_DESC(
  IncludeDiagSensors,
  "If nonzero, exposes certain register fields on certain models etc. as sensors, for diagnostic purposes. Default is 0.");

static uint LegacyHardware = 0;
module_param(LegacyHardware, uint, S_IRUGO);
MODULE_PARM_DESC(
  LegacyHardware,
  "If nonzero, includes hardware IDs of legacy hardware when searching for devices to control. Default is 0.");

static uint PciAddress64Bit = 0;
module_param(PciAddress64Bit, uint, S_IRUGO);
MODULE_PARM_DESC(
  PciAddress64Bit,
  "If nonzero, the driver may use 64-bit PCI addressing (if certain conditions are met) when programming bus master devices for DMA transfers. Default is 0.");

static uint Si5338ExposeAllClocks = 0;
module_param(Si5338ExposeAllClocks, uint, S_IRUGO);
MODULE_PARM_DESC(
  Si5338ExposeAllClocks,
  "If nonzero, clocks outputs from the SI5338 that are not intended to be user-programmable are exposed as being programmable. Default is 0.");

static uint PrimaryCommonBufferCount = 0;
module_param(PrimaryCommonBufferCount, uint, S_IRUGO);
MODULE_PARM_DESC(
  PrimaryCommonBufferCount,
  "Number of common buffers to allocate when the driver starts. Default is 0.");

static uint PrimaryCommonBufferAlignment = 32U;
module_param(PrimaryCommonBufferAlignment, uint, S_IRUGO);
MODULE_PARM_DESC(
  PrimaryCommonBufferCount,
  "Minimum alignment, in bytes, of common buffers allocated when the driver starts. Default is 32.");

static uint PrimaryCommonBufferSizeLow = 0x10000U;
module_param(PrimaryCommonBufferSizeLow, uint, S_IRUGO);
MODULE_PARM_DESC(
  PrimaryCommonBufferSizeLow,
  "Low 32 bits of size (in bytes) of common buffers allocated when the driver starts. Default is 0x10000.");

static uint PrimaryCommonBufferSizeHigh = 0;
module_param(PrimaryCommonBufferSizeHigh, uint, S_IRUGO);
MODULE_PARM_DESC(
  PrimaryCommonBufferSizeHigh,
  "High 32 bits of size (in bytes) of common buffers allocated when the driver starts. Default is 0.");

static uint PolledInterrupts = 0;
module_param(PolledInterrupts, uint, S_IRUGO);
MODULE_PARM_DESC(
  PolledInterrupts,
  "If nonzero, does not use interrupts, and the user-space \"polldaemon\" must be run in order to service interrupts. Default is 0.");

static uint DmaMptlReadBurstLength = 0;
module_param(DmaMptlReadBurstLength, uint, S_IRUGO);
MODULE_PARM_DESC(
  DmaMptlReadBurstLength,
  "Maximum burst length (in words) for OCP read commands issued by a DMA engine; 0 means use hardware default. Default is 0.");

static uint DmaMptlWriteBurstLength = 0;
module_param(DmaMptlWriteBurstLength, uint, S_IRUGO);
MODULE_PARM_DESC(
  DmaMptlWriteBurstLength,
  "Maximum burst length (in words) for OCP write commands issued by a DMA engine; 0 means use hardware default. Default is 0.");

DfLinuxDriverParameterUint32 g_dfLinuxDriverParamUint32[] = {
  { "Admxrc6tadv8CompatSensor1Name", &Admxrc6tadv8CompatSensor1Name },
  { "DebugLevel", &DebugLevel },
  { "DmaMapLength", &DmaMapLength },
  { "PrimaryCommonBufferCount", &PrimaryCommonBufferCount },
  { "PrimaryCommonBufferAlignment", &PrimaryCommonBufferAlignment },
  { "PrimaryCommonBufferSizeLow", &PrimaryCommonBufferSizeLow },
  { "DmaMptlReadBurstLength", &DmaMptlReadBurstLength },
  { "DmaMptlWriteBurstLength", &DmaMptlWriteBurstLength },
  { NULL, NULL } /* Terminator */
};
  
DfLinuxDriverParameterBoolean g_dfLinuxDriverParamBoolean[] = {
  { "AssumeXrmPresent", &AssumeXrmPresent },
  { "EnableVpdWrite", &EnableVpdWrite },
  { "IncludeDiagSensors", &IncludeDiagSensors },
  { "LegacyHardware", &LegacyHardware },
  { "PciAddress64Bit", &PciAddress64Bit },
  { "Si5338ExposeAllClocks", &Si5338ExposeAllClocks },
  { "PolledInterrupts", &PolledInterrupts },
  { NULL, NULL } /* Terminator */
};
  
DfLinuxDriverParameterString g_dfLinuxDriverParamString[] = {
  { NULL, NULL } /* Terminator */
};
  
/*
** PCI ID table used by hotplug subsystem. The table is in two parts, where the second part is for legacy ID.
** By putting NULL ID after the first part of the table, legacy hardware support is NOT advertised to the
** hotplug subsystem (intentional).
**
** If TRUE is passed in bSupportLegacyHardware, dfLinuxModuleInitPci() function removes the NULL ID after the
** first part and shifts the second part entries down, so that legacy hardware will be registered when the driver
** calls pci_register_driver.
**
** !!!NOTE!!! If there is no legacy hardware, the table MUST end with two NULL PCI ID records, e.g.
**
** static struct pci_device_id g_pciIdTable[] = {
**   { .vendor = ..., .device = ..., .subvendor = ..., .subdevice = ... },
**   { .vendor = ..., .device = ..., .subvendor = ..., .subdevice = ... },
**   { .vendor = 0x0U, .device = 0x0U },
**   { .vendor = 0x0U, .device = 0x0U }
** }
**
** In other words, if there is no legacy hardware, then do not remove the NULL ID for terminating the legacy
** hardware part of the table.
*/
static struct pci_device_id g_pciIdTable[] = {
  { .vendor = 0x4144U, .device = 0xadb3U, .subvendor = 0x4144U, .subdevice = PCI_ANY_ID }, /* ADB3-based models */
  { .vendor = 0x0U, .device = 0x0U }, /* Marks end of non-legacy hardware */
  /* Legacy hardware begins here */
  { .vendor = 0x10b5U, .device = 0x9656U, .subvendor = 0x4144U, .subdevice = PCI_ANY_ID }, /* PCI9656-based models */
  { .vendor = 0x0U, .device = 0x0U }  /* Marks end of table */
};

#if 0
/* When we have sorted out udev, we can enable the MODULE_DEVICE_TABLE so that the module loads automatically instead of via adb3.rc script */
MODULE_DEVICE_TABLE(pci, g_pciIdTable);
#endif

/*
** Module entry / exit points
*/

int __init
adb3Init(
  void)
{
#if DF_DBG_BUILD
  g_dfDebugLevel = (int)DebugLevel;
#endif
  return dfLinuxModuleInitPci(g_pciDriverName, g_pciIdTable, LegacyHardware ? TRUE : FALSE);
}

void __exit
adb3Exit(
  void)
{
  dfLinuxModuleExitPci();

#if DF_DBG_ALLOCATION
  dfLinuxReportLeaks();
#endif
}

module_init(adb3Init)
module_exit(adb3Exit)
